//====================================================================================//
//                                                                                    //
//                                     Gen_HBF                                        //
//                                                                                    //
//====================================================================================//
//  This File:   gen_hbf.C                   Language: C++  (xlC and CC)              //
//  Software:    PEPS 2006                   Advisor: Brigitte.Plateau@imag.fr        //
//====================================================================================//
//  Creation:    19/dec/95                        by: Paulo.Fernandes@imag.fr         //
//  Last Change: 13/feb/03                        by: Anne.Benoit@imag.fr         //
//====================================================================================//
#include "peps.h"

peps_name filename;

//====================================================================================//
void Generate_hbf()
//====================================================================================//
{                                // SAN model name in all directories handled by PEPS //
 peps_name model_name;     // current directory (just the name, no paths, no extensions)
 peps_name san_level_name; // directory where the .san is
 peps_name des_level_name; // directory where the textual descriptor is
 peps_name dsc_level_name; // directory where the internal format descriptor is
 peps_name matrix_name;    

 Timer     T;          // timer to inform total time spent

 T.clear();
 
 if(!Provide_DSC(filename))
  exit(0);
  Ask_a_File_Name(matrix_name, hbf_file);                          // .hbf generation
 if (Permission_to_Proceed(matrix_name, hbf_file)) {
   cout << "\nGenerating the sparse matrix ...\n";
   T.start();
   HBF::hbf.generate(matrix_name);
   T.stop();
   Notify_Time_Spend(T, "compilation of a SAN model and sparse matrix generation\n");
 }
}

//====================================================================================//
void Convert_hbf()
//====================================================================================//
{
 peps_name original_name, new_name;

 if (!strcmp(filename,""))
   Ask_an_Existing_File_Name(original_name, hbf_file);
 else{
   strcpy(original_name,filename);
   if (!Exists_File(original_name, hbf_file)){
     cout << "This file can not be found!";
     exit(0);
   }
 }  
 HBF::hbf.read(original_name);
 cout << "The matrix '" << original_name << ".hbf' is in ";
 switch (HBF::hbf.first_row()) {
 case 0  : cout << "the PEPS format (first column index is 0)\n";
           cout << "Converting to MARCA format (first row index is 1)\n";
	   HBF::hbf.convert("MATX_hb", F_format);
           break;
 case 1  : cout << "the MARCA format (first row index is 1)\n";
           Ask_a_File_Name(new_name, hbf_file);
           if (Permission_to_Proceed(new_name, hbf_file)) {
             cout << "Converting to PEPS format (first column index is 0)\n";
	     HBF::hbf.convert(new_name, C_format);
           } break;
 default : cout << "a unknown format\n ****** Conversion aborted! ****** \n\n";
 }
}


//====================================================================================//
//====================================================================================//

//====================================================================================//
//                                    Gen_HBF Main                                    //
//====================================================================================//
int main(int argc, char *argv[])
{
 cout.setf(ios::scientific);
 cout.precision(16);
 set_new_handler(Communicate_Out_of_Memory);
 Create_Temporary_Directories();
 int gen_op = nothing; 
 
 strcpy(filename,"");
 
 if (argc >= 2){
  for (int i=1; i<argc; i++){
    if (argv[i][0] == '-'){
      if (!(strcmp(argv[i], "-conv"))){
        if (gen_op == nothing)
          gen_op = convert_hbf;
         else
           cout << "warning: generation option is already set. Ignoring option \"-conv\"." << endl;
      }
      if (strcmp(argv[i], "-conv")){
        cout << "error: " << argv[i] << " invalid option" << endl; 
        exit (0);
      }
    }
    else{
      if (!strcmp(filename, ""))
        strcpy(filename, argv[i]);
      else
        cout << "warning: model name is already set as " << filename << ". Ignoring " << argv[i] << endl; 
    }
  }
 }
 

 switch (gen_op) {

//====================================================================================//
//                                    HBF Operations                                  //
//====================================================================================//
   case nothing      : 
   case generate_hbf : Generate_hbf();                            break;
   case convert_hbf  : Convert_hbf();                             break;

//====================================================================================//
//                                   Other Operations                                 //
//====================================================================================//
   case credits     : Show_Credits();                              break;
   default          : Programming_Error("invalid PEPS option", 0000)
 }
 Say_Goodbye();
 return(0);
}
